#include "task_manager.h"

#include "core/queue_manager.h"
#include "core/sync_update_sender.h"

#include "sync/update_api.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <unistd.h>

#include <cstdlib>
#include <cstdio>
#include <cstdarg>

#include <vector>
#include <algorithm>

using namespace std;
using namespace rcmdsync;

tr1::shared_ptr<TaskManager> TaskManager::task_manager(new TaskManager());

TaskManager & TaskManager::instance() {
    return *TaskManager::task_manager.get();
}

bool TaskManager::start(string const & _log_file_name, int _log_level) {
    assert(stopped);
    if (!stopped)
        return false;
    stopped = false;
    log_level = _log_level;
    log_file_name = _log_file_name;
    int fdres = open(log_file_name.c_str(), O_CREAT | O_APPEND | O_LARGEFILE | O_WRONLY, S_IWUSR | S_IRUSR | S_IRGRP | S_IROTH);
    log_file = fdres > 0 ? fdopen(fdres, "a") : NULL;
    return pthread_create(&thread, NULL, thread_func, this) == 0;
}

bool TaskManager::stop() {
    assert(!stopped);
    if (stopped)
        return false;
    {
        simple_lock(queue_lock);
        stopped = true;
        condition.signal();
    }
    bool result = pthread_join(thread, NULL) == 0;
    if (log_file != NULL)
        fclose(log_file);
    return result;
}

void * TaskManager::thread_func(void * param) {
    TaskManager * task_manager = reinterpret_cast<TaskManager *>(param);
    return reinterpret_cast<void *>(task_manager->run());
}

int TaskManager::run() {
    init_log_file(log_file_name.c_str(), 0);

    errlog(INF, "Task manager thread was started.");

    while(true) {
        tid_t tid;
        {
            simple_lock(queue_lock);
            while (true) {
                if (stopped)
                    break;
                if (!queue.empty())
                    break;
                condition.wait();
            }
            if (stopped)
                break;
            tid = queue.front();
            queue.pop();
        }

        uint16_t runner_id;
        set<tid_t> blocker;
        {
            simple_lock(tasks_lock);
            map<tid_t, task_t>::iterator task_it = tasks.find(tid);
            if (task_it == tasks.end()) // таск успел завершиться
                continue;
            runner_id = task_it->second.runner_id;
            blocker = task_it->second.blocker;
            errlog(DBG, "Selected task (%u): %s sid=%hu gsid=%u gtoid=%hu toid=%hu", static_cast<uint32_t>(tid), task_it->second.name.c_str(), static_cast<uint16_t>(task_it->second.sid), static_cast<uint16_t>(task_it->second.gsid), static_cast<uint16_t>(task_it->second.gtoid), static_cast<uint16_t>(task_it->second.toid));
        }


        if (runner_id == 0) { //  таск стоит в очереди на запуск
            errlog(DBG, "Task (%u) is going to run.", static_cast<uint32_t>(tid));
            if (!blocker.empty()) { // опрашиваем состояние блокирующих тасков
                errlog(DBG, "Task (%u) is blocked.", static_cast<uint32_t>(tid));
                simple_lock(queue_lock);
                for (set<tid_t>::iterator blocker_it = blocker.begin(), blocker_end = blocker.end(); blocker_it != blocker_end; ++blocker_it)
                    queue.push(*blocker_it);
                continue;
            }
            vector<pair<uint16_t, tr1::shared_ptr<SyncQueueSender> > > senders;
            {
                simple_lock(senders_lock);
                senders.reserve(TaskManager::senders.size());
                senders.insert(senders.end(), TaskManager::senders.begin(), TaskManager::senders.end());
            }
            bool started = false;
            for (size_t i = 0; i < senders.size(); ++i) {
                swap(senders[i], senders[i + rand() % (senders.size() - i)]);
                uint16_t id = senders[i].first;
                SyncUpdateSender * sender = dynamic_cast<SyncUpdateSender *>(senders[i].second.get());

                task_manager_info_t info;
                if (!get_info(info, sender))
                    continue;
                if (info.queue_size > 0)
                    continue;
                if (!start(tid, id, sender))
                    continue;
                started = true;
                break;
            }
            if (!started) {
                errlog(DBG, "Task (%u) wasn't started.", static_cast<uint32_t>(tid));
                {
                    simple_lock(queue_lock);
                    queue.push(tid);
                }
                usleep(10000);
            }
        } else { // таск выполняется на удалённой машине и требует проверки
            errlog(DBG, "Task (%u) is going to be checked.", static_cast<uint32_t>(tid));
            tr1::shared_ptr<SyncQueueSender> sender;
            {
                simple_lock(senders_lock);
                Senders::iterator sender_it = senders.find(runner_id);
                if (sender_it != senders.end())
                    sender = sender_it->second;
            }

            task_manager_info_t info;
            if (sender.get() == NULL || // клиент отключился
                !get_info(info, dynamic_cast<SyncUpdateSender *>(sender.get())) || // клиент упал
                info.task_list.find(tid) == info.task_list.end()) { // таск пропал на клиенте (таск был поставлен в очередь, успел завершится, или клент перезапустили)
                {
                    simple_lock(tasks_lock);
                    map<tid_t, task_t>::iterator task_it = tasks.find(tid);
                    if (task_it != tasks.end()) { // таск не был завершён. ставим в очередь
                        task_it->second.runner_id = 0;
                        task_it->second.flags |= task_it->second.running_flags;
                    }
                }
                {
                    simple_lock(queue_lock);
                    queue.push(tid);
                }
            }
        }

    }

    errlog(INF, "Task manager thread was finished.");

    destroy_log_file();
    return 0;
}

bool TaskManager::start(tid_t tid, uint16_t runner_id, rcmdsync::SyncUpdateSender * sender) {
    task_manager_start_in_t start_in;
    task_manager_start_out_t start_out;
    start_in.tid = tid;

    {
        simple_lock(tasks_lock);
        map<tid_t, task_t>::iterator task_it = tasks.find(tid);
        if (task_it == tasks.end())
            return false;
        assert(task_ids[task_it->second] == tid);

        start_in.name = task_it->second.name;
        start_in.sid = task_it->second.sid;
        start_in.gsid = task_it->second.gsid;
        start_in.gtoid = task_it->second.gtoid;
        start_in.toid = task_it->second.toid;
        start_in.flags = task_it->second.flags;

        task_it->second.runner_id = runner_id;
        task_it->second.running_flags = task_it->second.flags;
        task_it->second.flags = TASK_MANAGER_FLAG_NOTHING;
    }

    // Тут атомарность достигается за счёт того, что только один поток вызывает start

    if (!sender->send(TASK_MANAGER_START_API_ID, start_in, start_out) || !start_out.success) {
        simple_lock(tasks_lock);
        map<tid_t, task_t>::iterator task_it = tasks.find(tid);
        if (task_it == tasks.end())
            return false;
        task_it->second.runner_id = 0;
        task_it->second.flags |= task_it->second.running_flags;
        return false;
    }
    errlog(INF, "Task %u was started.", static_cast<uint32_t>(tid));
    return true;
}

bool TaskManager::get_info(task_manager_info_t & info, rcmdsync::SyncUpdateSender * sender) {
    task_manager_info_in_t info_in;
    task_manager_info_out_t & info_out = info;
    return sender->send(TASK_MANAGER_INFO_API_ID, info_in, info_out);
}

tid_t TaskManager::create(string const & name, sid_t sid, gsid_t gsid, gtoid_t gtoid, toid_t toid, set<tid_t> blockers, uint32_t flags) {
    errlog(INF, "Creating new task: %s (sid=%u, gsid=%u, gtoid=%u, toid=%u)", name.c_str(), static_cast<uint32_t>(sid), static_cast<uint32_t>(gsid), static_cast<uint32_t>(gtoid), static_cast<uint32_t>(toid));
	simple_lock(tasks_lock);
	// фильтруем завершённые таски
    for (set<tid_t>::const_iterator blocker_it = blockers.begin(), blocker_end = blockers.end(); blocker_it != blocker_end; ) {
        if (tasks.find(*blocker_it) != tasks.end())
            ++blocker_it;
        else
            blockers.erase(blocker_it++);
    }
    // Проверяем, существует ли уже нужный таск
	key_t key(name, sid, gsid, gtoid, toid);
	map<key_t, tid_t>::iterator task_id_it = task_ids.find(key);
	if (task_id_it != task_ids.end()) { // обновляем существующий таск
        tid_t & tid = task_id_it->second;
        map<tid_t, task_t>::iterator task_it = tasks.find(tid);
        assert(task_it != tasks.end());
        task_it->second.blocker.insert(blockers.begin(), blockers.end());
        task_it->second.flags |= flags;
	} else { // создаём новый таск
	    while (tasks.find(last_task_id) != tasks.end()) // тот редкий случай, когда таск пережил 4 миллиарда своих собратьев
	        ++last_task_id;
	    task_id_it = task_ids.insert(make_pair(key, last_task_id)).first;
        tasks.insert(make_pair(last_task_id, task_t(flags, blockers.begin(), blockers.end(), key)));
        ++last_task_id;
        if (last_task_id == 0) // на случай переполнения O_o
            last_task_id = 1;
	}
    tid_t & tid = task_id_it->second;
    // Добавляем в новый зависимости новый таск
    for (set<tid_t>::const_iterator blocker_it = blockers.begin(), blocker_end = blockers.end(); blocker_it != blocker_end; ++blocker_it) {
        map<tid_t, task_t>::iterator task_it = tasks.find(*blocker_it);
        if (task_it == tasks.end())
            continue;
        task_it->second.blocked.insert(tid);
    }
    // заносим в очередь на запуск или проверку
    {
        simple_lock(queue_lock);
        queue.push(tid);
        condition.signal();
    }
    errlog(DBG, "Task id: %u", static_cast<uint32_t>(tid));
	return tid;
}

bool TaskManager::finish(tid_t tid) {
    errlog(INF, "Task %u was finished.", static_cast<uint32_t>(tid));
    bool restart;
    set<tid_t> blocked;
    {
    	simple_lock(tasks_lock);
	    map<tid_t, task_t>::iterator task_it = tasks.find(tid);
    	if (task_it == tasks.end())
	    	return false;
    	assert(task_ids[task_it->second] == tid);
    	restart = task_it->second.flags != TASK_MANAGER_FLAG_NOTHING;
    	if (restart) {
    	    task_it->second.runner_id = 0;
    	} else { // Если за время работы таска кто-то запрашивал таск, то перезапускаем
            blocked = task_it->second.blocked;
            task_ids.erase(task_it->second);
            tasks.erase(task_it);
    	}
    }

    if (restart) {
        errlog(INF, "Task %u was restarted.", static_cast<uint32_t>(tid));
        simple_lock(queue_lock);
        queue.push(tid);
        condition.signal();
    }

	for (set<tid_t>::iterator blocked_it = blocked.begin(), blocked_end = blocked.end(); blocked_it != blocked_end; ++blocked_it) {
	    tid_t const & blocked_tid = *blocked_it;
        {
            simple_lock(tasks_lock);
            map<tid_t, task_t>::iterator blocked_task_it = tasks.find(blocked_tid);
	        assert(blocked_task_it != tasks.end());
    	    blocked_task_it->second.blocker.erase(tid);
        }
        {
            simple_lock(queue_lock);
            queue.push(blocked_tid);
        	condition.signal();
        }
	}
	return true;
}

std::vector<task_manager_record_t> TaskManager::list() {
    errlog(INF, "Listing tasks");
	simple_lock(tasks_lock);
	vector<task_manager_record_t> result;
	result.reserve(tasks.size());
	for (map<tid_t, task_t>::const_iterator task_it = tasks.begin(), task_end = tasks.end(); task_it != task_end; ++task_it) {
		task_manager_record_t record;
		record.tid = task_it->first;
		record.name = task_it->second.name;
		record.sid = task_it->second.sid;
		record.gsid = task_it->second.gsid;
		record.gtoid = task_it->second.gtoid;
		record.toid = task_it->second.toid;
		result.push_back(record);
	}
	return result;
}

bool TaskManager::subscribe(uint16_t id, std::tr1::shared_ptr<rcmdsync::SyncQueueSender> const & sender) {
    simple_lock(senders_lock);
    return senders.insert(make_pair(id, sender)).second;
}

bool TaskManager::unsubscribe(uint16_t id) {
    simple_lock(senders_lock);
    return senders.erase(id) != 0;
}

void TaskManager::errlog(int level, char const * format, ...) {
    if (log_file == NULL)
        return;
    if (!(log_level & level))
        return;
    simple_lock(log_lock);
    write_time(log_file);
    fprintf(log_file, "[%s]: ", level_name[level]);
    va_list arg;
    va_start(arg, format);
    vfprintf(log_file, format, arg);
    va_end(arg);
    fprintf(log_file, "\n");
    fflush(log_file);
}
